"""
This module provides an interface to the FingerWorks iGesture pads,
a multitouch device that was sold all-to-briefly by Fingerworks.
See http://www.fingerworks.com for more info.
"""

import sys
import time
import traceback
import thread
import threading
import win32con
import igesture
import copy

from threading import Thread,Lock
from math import sqrt
from ctypes import *

class BaseEvent:
	def __init__(self):
		self.time = 0.0

WNDPROC = WINFUNCTYPE(c_long, c_int, c_uint, c_int, c_int)

FINGER_DOWN = 0
FINGER_UP = 1
FINGER_DRAG = 2

class WNDCLASS(Structure):
	_fields_ = [('style', c_uint),
				('lpfnWndProc', WNDPROC),
				('cbClsExtra', c_int),
				('cbWndExtra', c_int),
				('hInstance', c_int),
				('hIcon', c_int),
				('hCursor', c_int),
				('hbrBackground', c_int),
				('lpszMenuName', c_char_p),
				('lpszClassName', c_char_p)]

class RECT(Structure):
	_fields_ = [('left', c_long),
				('top', c_long),
				('right', c_long),
				('bottom', c_long)]

class PAINTSTRUCT(Structure):
	_fields_ = [('hdc', c_int),
				('fErase', c_int),
				('rcPaint', RECT),
				('fRestore', c_int),
				('fIncUpdate', c_int),
				('rgbReserved', c_char * 32)]

class POINT(Structure):
	_fields_ = [('x', c_long),
				('y', c_long)]
	
class MSG(Structure):
	_fields_ = [('hwnd', c_int),
				('message', c_uint),
				('wParam', c_int),
				('lParam', c_int),
				('time', c_int),
				('pt', POINT)]

def ErrorIfZero(handle):
	if handle == 0:
		raise WinError
	else:
		return handle

def formatExceptionInfo(maxTBlevel=5):
	cla, exc, trbk = sys.exc_info()
	excName = cla.__name__
	try:
		excArgs = exc.__dict__["args"]
	except KeyError:
		excArgs = "<no args>"
	excTb = traceback.format_tb(trbk, maxTBlevel)
	return (excName, excArgs, excTb)


def MainWin(g):
	CreateWindowEx = windll.user32.CreateWindowExA
	CreateWindowEx.argtypes = [c_int, c_char_p, c_char_p, c_int, c_int, c_int, c_int, c_int, c_int, c_int, c_int, c_int]
	CreateWindowEx.restype = ErrorIfZero

	# Define Window Class
	wndclass = WNDCLASS()
	wndclass.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
	wndclass.lpfnWndProc = WNDPROC(WndProc)
	wndclass.cbClsExtra = wndclass.cbWndExtra = 0
	wndclass.hInstance = windll.kernel32.GetModuleHandleA(c_int(win32con.NULL))
	wndclass.hIcon = windll.user32.LoadIconA(c_int(win32con.NULL), c_int(win32con.IDI_APPLICATION))
	wndclass.hCursor = windll.user32.LoadCursorA(c_int(win32con.NULL), c_int(win32con.IDC_ARROW))
	wndclass.hbrBackground = windll.gdi32.GetStockObject(c_int(win32con.WHITE_BRUSH))
	wndclass.lpszMenuName = None
	wndclass.lpszClassName = "MainWin"
	# Register Window Class
	if not windll.user32.RegisterClassA(byref(wndclass)):
		raise WinError()
	# Create Window
	hwnd = CreateWindowEx(0,
		  wndclass.lpszClassName,
		  "Python Window",
		  win32con.WS_OVERLAPPEDWINDOW,
		  win32con.CW_USEDEFAULT,
		  win32con.CW_USEDEFAULT,
		  win32con.CW_USEDEFAULT,
		  win32con.CW_USEDEFAULT,
		  win32con.NULL,
		  win32con.NULL,
		  wndclass.hInstance,
		  win32con.NULL)

	g.init(hwnd)

	g.start_callbacks()

	# Show Window
	# windll.user32.ShowWindow(c_int(hwnd), c_int(win32con.SW_SHOWNORMAL))
	windll.user32.ShowWindow(c_int(hwnd), c_int(win32con.SW_HIDE))
	windll.user32.UpdateWindow(c_int(hwnd))
	# Pump Messages
	msg = MSG()
	pMsg = pointer(msg)
	NULL = c_int(win32con.NULL)
	
	while windll.user32.GetMessageA( pMsg, NULL, 0, 0) != 0:
		windll.user32.TranslateMessage(pMsg)
		windll.user32.DispatchMessageA(pMsg)
	g.processframes()

	return msg.wParam
	
def WndProc(hwnd, message, wParam, lParam):

	ps = PAINTSTRUCT()
	rect = RECT()
	
	if message == win32con.WM_PAINT:
		hdc = windll.user32.BeginPaint(c_int(hwnd), byref(ps))
		windll.user32.GetClientRect(c_int(hwnd), byref(rect))
		windll.user32.DrawTextA(c_int(hdc),
			"Python Powered Windows" ,
			c_int(-1), byref(rect), 
			win32con.DT_SINGLELINE|win32con.DT_CENTER|win32con.DT_VCENTER)
		windll.user32.EndPaint(c_int(hwnd), byref(ps))
		return 0
	elif message == win32con.WM_DESTROY:
		windll.user32.PostQuitMessage(0)
		return 0

	return windll.user32.DefWindowProcA(c_int(hwnd), c_int(message), c_int(wParam), c_int(lParam))

class Finger:

	oneThread = None

	@staticmethod
	def startup():
		if Finger.oneThread:
			raise Exception, "Finger has already been started"
		Finger.oneThread = FingerThread()
		Finger.oneThread.start()

	@staticmethod
	def num_devices():
		if Finger.oneThread == None:
			raise Exception, "Finger hasn't been started"
		return Finger.oneThread.num_devices()

	@staticmethod
	def time_now():
		return time.time()

	@staticmethod
	def callback(f,data):
		if Finger.oneThread == None:
			raise Exception, "Finger hasn't been started"
		return Finger.oneThread.callback(f,data)

class FingerDevice:

	def __init__(self,d):
		if Finger.oneThread == None:
			raise Exception, "Finger hasn't been started"
		self.width = Finger.oneThread.device_width(d)
		self.height = Finger.oneThread.device_height(d)
		self.devindex = d

	def to_xml(self):
		return '<igesture_pad devindex="%d" width="%f" height="%f"/>' % (self.devindex,self.width,self.height)


class FingerState:
	"""
	Provides accumulated state for tracking fingers.
	"""

	def __init__(self,d,f):
		self.devindex = d
		self.finger = f
		self.x = 0.0
		self.y = 0.0
		self.prox = 0.0

	def __str__(self):
		return "FingerState(d=%d f=%d x=%.3f y=%.3f prox=%.3f)" % (self.devindex,self.finger,self.x,self.y,self.prox)

class FingerEvent(BaseEvent):
	"""
	Provides history of finger data
	"""

	def __init__(self,f):
		BaseEvent.__init__(self)
		self.devindex = f.devindex
		self.finger = f.finger
		self.x = f.x
		self.y = f.y
		self.prox = f.prox
		self.time = f.time
		self.uid = f.uid

	def __str__(self):
		return "FingerEvent(type=\"%s\" time=%.3f devindex=%d finger=%d x=%.3f y=%.3f prox=%.3f)" % (self.type_string(),self.time,self.devindex,self.finger,self.x,self.y,self.prox)

	def to_xml(self):
		return '<event time="%.3f"><finger_%s devindex="%d" finger="%d" x="%.3f" y="%.3f" prox="%.3f" /></event>' % (self.time,self.type_string(),self.devindex,self.finger,self.x,self.y,self.prox)

class FingerUp(FingerEvent):

	def __init__(self,f):
		FingerEvent.__init__(self,f)

	def type_string(self):
		return "up"

class FingerDown(FingerEvent):

	def __init__(self,f):
		FingerEvent.__init__(self,f)

	def type_string(self):
		return "down"

class FingerDrag(FingerEvent):

	def __init__(self,f):
		FingerEvent.__init__(self,f)

	def type_string(self):
		return "drag"

class FingerThread(Thread):
	"""
	Provides cooked access to the igesture interface, keeping
	track of what fingers are currently down, smoothing drag motion, etc.
	"""

	def __init__(self):
		Thread.__init__(self)
		self.setDaemon(True)
		self.g = igesture.device()
		self.infinger = 0
		self.fingers = {}
		self.fingxoff = 0.0
		self.fingyoff = 0.0
		self.fingerdevices = {}
		self.debug = False
		self.callback_func = None
		self.callback_data = None

	def run(self):
		self.g.setcallback(self.gotfinger)
		return MainWin(self.g)

	def gotfinger(self,fdev,fnum,event,x,y,prox):

		if not (fdev in self.fingerdevices):

			# We want to guarantee that the first finger device
			# touched is numbered 0, the second is 1, etc.
			# d = len(self.fingerdevices)
			# self.fingerdevices[fdev] = d

			# NEW STRATEGY - keep the original device numbers
			d = fdev
			self.fingerdevices[fdev] = d

		else:
			d = self.fingerdevices[fdev]

		# At this point, d is the device number (starting at 0)

		# For some reason, x and y are sometimes offset,
		# so make it self-adjusting to shift the coordinate space
		# and make sure that it goes from 0.0 to 1.0
		x += self.fingxoff
		y += self.fingyoff

		if x < 0.0:
			self.fingxoff -= x
			x = 0.0
		elif x > 1.0:
			self.fingxoff -= (x-1.0)
			x = 1.0

		if y < 0.0:
			self.fingyoff -= y
			y = 0.0
		elif y > 1.0:
			self.fingyoff -= (y-1.0)
			y = 1.0

		if self.infinger:
			if self.debug:
				print("GOTFINGER skipped due to recursion!")
			return
		self.infinger = 1
		i = d * 100 + fnum

		if not i in self.fingers:
			self.fingers[i] = FingerState(d,fnum)

		fing = self.fingers[i]
		fing.uid = i
		fing.x = x
		fing.y = y
		fing.prox = prox
		fing.time = Finger.time_now()

		try:
			if event == FINGER_DOWN:

				# Give some time to other threads
				time.sleep(0.001)

				fing.speed = 0.1
				fing.lastndx = 0.0
				fing.lastndy = 0.0
				self.push_event(FingerDown(fing))

			elif event == FINGER_DRAG:

				self.dofingerdrag(fing)

			elif event == FINGER_UP:

				# Give some time to other threads
				time.sleep(0.001)
				self.push_event(FingerUp(fing))
				del self.fingers[i]
				fing = None
		except:
			print formatExceptionInfo()

		if fing != None:
			fing.prevx = fing.x
			fing.prevy = fing.y
			fing.prevprox = fing.prox
			fing.prevtime = fing.time

		self.infinger = 0

	def callback(self,f,data=None):
		self.callback_func = f
		self.callback_data = data

	def push_event(self,e):
		if self.callback_func:
			self.callback_func(e,self.callback_data)

	def dofingerdrag(self,fing):

		# This is where smoothing of the finger
		# values gets done.
		# Also where the angle/direction of
		# finger movement gets done.

		dx = fing.x - fing.prevx
		dy = fing.y - fing.prevy
		dprox = fing.prox - fing.prevprox
		lng = sqrt(dx**2+dy**2)
		dtime = fing.time - fing.prevtime
		
		# No change in position or time, do nothing
		if lng == 0.0 or dtime == 0.0:
			# print "NO finger/time CHANGE, doing nothing"
			return

		# print "dx=",dx," dy=",dy," lng=",lng," dtime=",dtime
		thisspeed = lng/dtime

		speedadjust = False
		if speedadjust:
			accellimit = 1.4
			speedlimit = fing.speed * accellimit
			# print "thisspeed=",thisspeed," lastspeed=",fing.speed
			if thisspeed > speedlimit:
				print "ACC LIMIT!"
				thisspeed = speedlimit
			abslimit = 3.0
			if thisspeed > abslimit:
				print "ABS LIMIT!"
				thisspeed = abslimit

		fing.speed = thisspeed
		ndx = fing.speed * dtime * (dx/lng)
		ndy = fing.speed * dtime * (dy/lng)

		damp = 3.0
		ndx = (damp*fing.lastndx + ndx)/(damp+1.0)
		ndy = (damp*fing.lastndy + ndy)/(damp+1.0)
		# print "lastndxy=",fing.lastndx,fing.lastndy," ndxy=",ndx,ndy
		nprox = (damp*fing.prevprox + fing.prox)/(damp+1.0)

		nx = fing.prevx + ndx
		ny = fing.prevy + ndy
		fing.lastndx = ndx
		fing.lastndy = ndy
		fing.x = nx
		fing.y = ny
		fing.prox = nprox
		self.push_event(FingerDrag(fing))

	def num_devices(self):
		return self.g.numdevices()

	def device_width(self,d):
		return self.g.device_width(d)

	def device_height(self,d):
		return self.g.device_height(d)

	def get_finger(self,fi):
		return self.fingers[fi]

	def get_fingers(self):
		return copy.copy(self.fingers)

