#include <windows.h>
#include "Python.h"
#include "../fingerworks/include/FWHID_MultiTouchDevice.h"
#include "igesturelib.h"

#define GESTURE_MAX_DEVICES 10
#define GESTURE_MAX_FINGERS 6

struct fingdevice_s;

typedef enum fingereventtype_e {
	FINGER_DOWN = 0,
	FINGER_UP = 1,
	FINGER_DRAG = 2
} fingereventtype;

typedef struct finger_s {
	int isdown;
	float xpos;
	float ypos;
	int hand_id;
	float xvel;
	float yvel;
	float proximity;
	float orientation;
	float eccentricity;
	int devnum;
	int fingnum;
	long up_timeout;
} finger_t;

typedef struct fingdevice_s {
	float width_cm;
	float height_cm;
	float xoff;
	float yoff;
	char *name;
	finger_t finger[GESTURE_MAX_FINGERS];
	char touched[GESTURE_MAX_FINGERS];
	int processed;
	int discovered;	/* discovery order */
} fingdevice_t;

static int Nfingdevices = 0;
static fingdevice_t Fingdevice[GESTURE_MAX_DEVICES];
static FWHID_ContactFrame ContactFrame[GESTURE_MAX_DEVICES];
static FWMultiTouchDevicePtr gDevice[GESTURE_MAX_DEVICES];
static int gInitialized = 0;
static int processframenested = 0;

HANDLE gMutex;
HWND gKhwnd;

static int Tcount = 0;
static int Discovered = 0;

static int
gesture_lockframedata(char *s)
{
	DWORD dwWaitResult;
	dwWaitResult = WaitForSingleObject(gMutex,5000);
	switch(dwWaitResult){
		case WAIT_OBJECT_0:
			return TRUE;
		case WAIT_TIMEOUT:
		default:
			return FALSE;
	}
}

static void
gesture_unlockframedata(void)
{
	ReleaseMutex(gMutex);
}

static FingerCallback Fcallback = NULL;

/*
 *  What comes into this function is centimeters, and what
 *  comes out is something thats scaled and guaranteed to be
 *  between 0.0 and 1.0
 */
static void
gesture_normalize(float *xposp, float *yposp, fingdevice_t *gd)
{
	float xpos = *xposp;
        float ypos = *yposp;

	xpos += gd->xoff;
	ypos += gd->yoff;
	/*
	 * Sometimes the gesture pads report values which are < 0 or 
	 * greater than the width, so we adjust for this on a per-pad basis.
	 */
	if ( xpos < 0.0 ) {
		gd->xoff += (-xpos);
		xpos = 0.0;
	} else if ( xpos > gd->width_cm ) {
		gd->xoff += (gd->width_cm - xpos);
		xpos = gd->width_cm;
	}
	if ( ypos < 0.0 ) {
		gd->yoff += (-ypos);
		ypos = 0.0;
	} else if ( ypos > gd->height_cm ) {
		gd->yoff += (gd->height_cm - ypos);
		ypos = gd->height_cm;
	}
	*xposp = xpos/gd->width_cm;
	*yposp = ypos/gd->height_cm;
}

//This is called after every frame
static void
DeviceFrameCompleteCallback(FWMultiTouchDevicePtr fwdevice, void *targetData)
{
	//Find which MultiTouch device this is
	int i=0;
	int locked;

	while(fwdevice != gDevice[i] && i < Nfingdevices)
		i++;

	if(i == Nfingdevices)
		return; //Device not found in tracking list

	locked = gesture_lockframedata("framecallback");

	FWHID_getContactFrame(gDevice[i],&(ContactFrame[i]));
	Fingdevice[i].processed = 0;

	if ( locked )
		gesture_unlockframedata();

}

//The Device Creation Callback is called for every FW device found in the system
static void 
DeviceCreationCallback(FWMultiTouchDevicePtr dev, void *targetData)
{
	FWSurfaceRectangle r;
	char *s;

	if(Nfingdevices < GESTURE_MAX_DEVICES){
		FWHID_disableStreams(dev,kFW_DefaultStreams);
		FWHID_enableStreams(dev,kFW_ContactPathStream);		
		gDevice[Nfingdevices] = dev;

		r = FWHID_getSurfaceDimensions(dev);
		Fingdevice[Nfingdevices].width_cm = (float)r.width_cm;
		Fingdevice[Nfingdevices].height_cm = (float)r.height_cm;
		Fingdevice[Nfingdevices].xoff = 0.0;
		Fingdevice[Nfingdevices].yoff = 0.0;
		Fingdevice[Nfingdevices].discovered = Discovered++;

		s = FWHID_getProductName(dev);
		if ( s == NULL )
			s = "NullReturnFromGetProductName";
		Fingdevice[Nfingdevices].name = s;

		FWHID_setContactFrameCallback(
			gDevice[Nfingdevices],DeviceFrameCompleteCallback);


		Nfingdevices++;
	}
}

//Called everytime a FW device is removed.
//Note: If a device is plugged in all devices are removed and then rediscovered
void 
DeviceDisposalCallback(FWMultiTouchDevicePtr fwdevice, void *targetData)
{
	int i=0;

	while(fwdevice != gDevice[i] && i < Nfingdevices)
		i++;

	if(i == Nfingdevices)
		return; //device removed was not found in devices

	FWHID_disableStreams(fwdevice,kFW_ContactPathStream);
	FWHID_enableStreams(fwdevice,kFW_DefaultStreams);

	if(i+1 < GESTURE_MAX_DEVICES)
		memmove(&gDevice[i],&gDevice[i+1],
			sizeof(FWMultiTouchDevicePtr)*(Nfingdevices-i+1));

	Nfingdevices--;
}

#define ignore_contact(cp) (((cp)->finger_id==0)||((cp)->proximity<0.001))

// NOTE: we assume the lock is held when this function is called
static void
gesture_processframe(int dev)
{
	int ci;
	int fi;
	fingdevice_t *devp;
	FWHID_Contact *cp;
	finger_t *fp;
	DWORD tm = GetTickCount();
	FWHID_ContactFrame *cframep;
	int locked;

	locked = gesture_lockframedata("processframes");

	cframep = &ContactFrame[dev];

	devp = &(Fingdevice[dev]);
	if ( cframep == NULL || devp->processed ) {
		if ( locked )
			gesture_unlockframedata();
		goto dotimeouts;
	}
	devp->processed = 1;

	memset(devp->touched,0,GESTURE_MAX_FINGERS);

	for ( fi=0; fi<GESTURE_MAX_FINGERS; fi++) {
		int ncontacts = cframep->contact_count;
		for ( ci=0; ci<ncontacts; ci++ ) {

			cp = &(cframep->contacts[ci]);
			if ( ignore_contact(cp) )
				continue;
			if ( (cp->finger_id-1) == fi ) {
				fp = &(devp->finger[fi]);
				fp->fingnum = fi;

				fp->xpos = cp->xpos;
				fp->ypos = cp->ypos;
				gesture_normalize(&(fp->xpos),&(fp->ypos),devp);

				fp->xpos = (float)(cp->xpos / devp->width_cm);
				fp->ypos = (float)(cp->ypos / devp->height_cm);
				fp->hand_id = cp->hand_id;
				fp->xvel = (float)cp->xvel;
				fp->yvel = (float)cp->yvel;
				fp->proximity = (float)cp->proximity;
				fp->orientation = (float)cp->orientation;
				fp->eccentricity = (float)cp->eccentricity;
				fp->devnum = dev;
				devp->touched[fi] = 1;
			}
		}
	}

	if ( locked )
		gesture_unlockframedata();

	// SHOULD LOCK FINGER DATA?  processframes is mutexed, though

	for ( fi=0; fi<GESTURE_MAX_FINGERS; fi++) {
		fp = &(devp->finger[fi]);
		if ( devp->touched[fi] == 0 ) {
			if ( fp->isdown ) {
				// SHOULDN'T DO THIS RIGHT AWAY!
				// WAIT FOR N MILLISECONDS BEFORE
				// CONCLUDING IT'S UP.
				if ( fp->up_timeout == 0 ) {
#define UP_TIMEOUT 100
					fp->up_timeout = tm+UP_TIMEOUT;
				}
			}
		} else {
			if ( fp->isdown ) {
			
				fp->up_timeout = 0;
				if ( Fcallback ) {
					(Fcallback)(fp->devnum,fp->fingnum,FINGER_DRAG,fp->xpos,fp->ypos,fp->proximity);
				}
			} else {
				fp->isdown = 1;
				fp->up_timeout = 0;
				if ( Fcallback ) {
					(Fcallback)(fp->devnum,fp->fingnum,FINGER_DOWN,fp->xpos,fp->ypos,fp->proximity);
				}
			}
		}
	}
dotimeouts:
	// See if any of the timeout ones are due
	// NOTE: might not want to do this all the time?
	for ( fi=0; fi<GESTURE_MAX_FINGERS; fi++) {
		fp = &(devp->finger[fi]);
		if ( fp->isdown 
			&& (fp->up_timeout > 0) 
			&& (fp->up_timeout < tm )) {

			fp->isdown = 0;
			if ( Fcallback ) {
				(Fcallback)(fp->devnum,fp->fingnum,FINGER_UP,fp->xpos,fp->ypos,fp->proximity);
			}
		}
	}
}

/****************************************************************
 *  Below are the public functions of this library.
 ****************************************************************/

int
igesturelib_init(int hwnd)
{
	// try this
	int dev;
	int fi;

	gMutex = CreateMutex(NULL, FALSE, NULL);
	if ( gMutex == NULL ) {
		return FALSE;
	}
	Nfingdevices = 0;
	memset(Fingdevice,0,sizeof(Fingdevice));

	for ( dev=0; dev<GESTURE_MAX_DEVICES; dev++ ) {
		for ( fi=0; fi<GESTURE_MAX_FINGERS; fi++) {
			Fingdevice[dev].finger[fi].fingnum = fi;
		}
		Fingdevice[dev].processed = 1;
	}
	FWHID_setDeviceCreationCallback(DeviceCreationCallback);
	FWHID_setDeviceDisposalCallback(DeviceDisposalCallback);
	FWHID_registerMultiTouchDevices((HWND)hwnd);
	gInitialized = 1;
	gKhwnd = (HWND)hwnd;
	return TRUE;
}

int
igesturelib_processframes()
{
	int i;

	// SHOULD LOCK AROUND PROCESSFRAMENESTED ACCESS?

	if ( processframenested > 0 ) {
		// printf("PROCESSFRAMENESTED = %d!!!\n",processframenested);
		return 0;
	}
	processframenested++;

	for ( i=0; i < Nfingdevices; i++ ) {
		gesture_processframe(i);
	}

	processframenested--;

	return 0;
}

int
igesturelib_setcallback(FingerCallback f)
{
	Fcallback = f;
	return 0;
}

int
igesturelib_end()
{
	if ( gInitialized ) {
		gInitialized = 0;
		FWHID_releaseMultiTouchDevices();
		return 0;
	}
	return 1;
}

int
igesturelib_numdevices()
{
	return Nfingdevices;
}

float
igesturelib_width(int n)
{
	if ( n >= Nfingdevices )
		return 0.0;
	return Fingdevice[n].width_cm;
}

float
igesturelib_height(int n)
{
	if ( n >= Nfingdevices )
		return 0.0;
	return Fingdevice[n].height_cm;
}

